import { randomInt } from '../../lib/utils/index.mjs';

/**
 * @function
 * @param {String} mac - hardware device MAC address
 * @param {String} sublocation_uid - sublocation unique identifier
 * @param {String} tk_uid - trackable unique id
 * @return {Object}
 */
export function createPosition(mac, sublocation_uid, tk_uid) {
    return {
        ts: new Date().toISOString(),
        tk_uid,
        data: {
            x: randomInt(),
            y: randomInt(),
            z: randomInt(),
            mac,
            type: 'TST',
            sl_uid: sublocation_uid,
            fom: randomInt() % 100
        }
    };
}

/**
 *
 * @param {String} mac - hardware device MAC address
 * @param {String} sublocation_uid - sublocation unique identifier
 * @param {String} tk_uid - trackable unique id
 * @returns {{plist: {sl_id, ntype: string}, sublocation_uid, ms: number, x: number, action: string, y: number, z: number, mac, ts: number, dev_id, object: string}}
 */
export function createHeartbeat(mac, sublocation_uid, tk_uid) {
    const date = new Date();
    const ts = Math.floor(date.getTime() / 1000);
    const ms = date.getTime() % 1000;
    return {
        ts,
        ms,
        dev_id: tk_uid,
        mac,
        object: 'TO',
        action: 'HB',
        x: randomInt(),
        y: randomInt(),
        z: randomInt(),
        plist: {
            sl_id: sublocation_uid,
            ntype: 'NAV'
        },
        sublocation_uid
    };
}

/**
 *
 * @param {Number} limit - position count to send
 * @param {String} mac - tag mac address
 * @param {String} sublocation_uid - sublocation unique id
 * @return {{next: (function(): {value: (undefined|{data: {fom: number, sl_uid: String, x: number, y: number, z: number, type: string, mac: String}, ts: string}), done: boolean})}}
 * @constructor
 */
function PositionIterator({ limit, mac, sublocation_uid, tk_uid }) {
    let index = 0;
    let done = false;
    return {
        // return the next method, need for iterator
        next: () => {
            done = (index >= limit);
            let obj = {
                value: done ? undefined : createPosition(mac, sublocation_uid, tk_uid),
                done
            };
            index++; // increment index
            return obj;
        }
    };
}

/**
 *
 * @param param
 * @property {String} mac - hardware device MAC address
 * @property {Number} limit - heartbeat count
 * @property {String} sublocation_uid - sublocation unique identifier
 * @property {String} tk_uid - trackable unique id
 * @returns {{next: (function(): {value: (undefined|{plist: {sl_id, ntype: string}, sublocation_uid, ms: number, x: number, action: string, y: number, z: number, mac, ts: number, dev_id, object: string}), done: boolean})}}
 * @constructor
 */
function HeartbeatIterator({ limit, mac, sublocation_uid, tk_uid }) {
    let index = 0;
    let done = false;
    return {
        // return the next method, need for iterator
        next: () => {
            done = (index >= limit);
            let obj = {
                value: done ? undefined : createHeartbeat(mac, sublocation_uid, tk_uid),
                done
            };
            index++; // increment index
            return obj;
        }
    };
}


/**
 *
 * @param param
 * @property {String} mac - hardware device MAC address
 * @property {Number} limit - position count
 * @property {String} sublocation_uid - sublocation unique identifier
 * @property {String} tk_uid - trackable unique id
 * @returns {{[Symbol.iterator]: (function(): {next: (function(): {value: (undefined|{data: {fom: number, sl_uid: String, x: number, y: number, z: number, type: string, mac: String}, ts: string}), done: boolean})})}}
 */
export const makePositionsIterator = ({ limit, mac, sublocation_uid, tk_uid }) => {
    return {
        [Symbol.iterator]: () => PositionIterator({ limit, mac, sublocation_uid, tk_uid })
    };
};

/**
 *
 * @param param
 * @property {String} mac - hardware device MAC address
 * @property {Number} limit - heartbeat count
 * @property {String} sublocation_uid - sublocation unique identifier
 * @property {String} tk_uid - trackable unique id
 * @returns {{[Symbol.iterator]: (function(): {next: function(): {value: (undefined|{plist: {sl_id, ntype: string}, sublocation_uid, ms: number, x: number, action: string, y: number, z: number, mac, ts: number, dev_id, object: string}), done: boolean}})}}
 */
export const makeHeartbeatIterator = (param) => {
    return {
        [Symbol.iterator]: () => HeartbeatIterator(param)
    };
};

export async function delay(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}
