import { SolutionDBPoolFactory } from '../db/solutionDbClient.mjs';
import { toMySqlTime, positionToString } from '../utils/index.mjs';
import logger from '../api/logger.mjs';


export async function delay(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}

export class PositionManager {
    positions = '';
    dbClient;
    timer;
    dbConnection;
    timeout;
    count = 0;
    restoreConnectionCount  = 0;
    
    /**
     *
     * @param options
     * @property options.solution_db - mariaDB connection configuration
     * @property options.solution_db.host
     * @property options.solution_db.user
     * @property options.solution_db.port
     * @property options.solution_db.password
     * @property options.solution_db.database
     * @property options.solution_db.driver
     * @property options.solution_db.charset
     * @property options.solution_db.collate
     * @property options.saveTimeout
     */
    constructor(options) {
        this.savePositions.bind(this);
        this.dbClient = SolutionDBPoolFactory.createSolutionDBPool(options.solution_db);
        this.timeout = options.saveTimeout;
        this.timer = setInterval(async () => {
            await this.savePositions();
        }, this.timeout || 1500);
    }

    /**
     * Check position validness
     * @param pos
     * @property {String} pos.tk_uid - trackable unique id
     * @property {String} pos.data.sl_uid - sublocation unique id
     * @property {String} pos.data.ts - timestamp in ISO 8601 format
     * @returns {boolean}
     */
    isPositionValid(pos) {
        return pos.tk_uid && pos.data.sl_uid && pos.ts;
    }
    /**
     *
     * @param {Object} pos
     * @property {String} pos.tk_uid - trackable unique id
     * @property {String} pos.ts - timestamp in ISO 8601 format
     * @property {Object} pos.data - position data
     * @property {Number} pos.data.x - trackable coordinate X
     * @property {Number} pos.data.y - trackable coordinate Y
     * @property {Number} pos.data.z - trackable coordinate Z
     * @property {String} pos.data.type - position type (UWB|TST)
     * @property {String} pos.data.sl_uid - sublocation unique id
     * @property {String} pos.data.mac - trackable device mac address
     * @return {void}
     */
    handlePosition(pos) {
        // this.count = this.count + 1;
        if (this.isPositionValid(pos)) {
            // logger.info('++++++  POS COUNT: ',this.count);
            // logger.info('pos: ', JSON.stringify(pos));
            pos.ts = toMySqlTime(pos.ts);
            if (this.positions.length > 0) {
                this.positions += ',';
            }
            this.positions += positionToString(pos);
        } else {
            // logger.warn('Position without trackable or sublocation: ', pos);
        }
    }


    /**
     * Save batch of positions every interval (see this.timeout)
     * @return {Promise<void>}
     */
    async savePositions() {
        if (!this.dbConnection || !this.dbConnection.isValid()) {
            this.dbConnection = await this.dbClient.pool.getConnection();
        }
        if (this.positions.length > 0) {
            const inserts = this.positions;
            this.positions = '';
            try {
                const sql = `INSERT INTO position_histories (obj_uid, mac_address, ts, x, y, z, sublocation_uid, type, fom)
                             VALUES ${inserts};`;
                await this.dbConnection.query(sql, inserts);
                this.restoreConnectionCount = 0;
            } catch (e) {
                this.positions = inserts;
                this.restoreConnectionCount++;
                logger.error('Error inserting positions, reverting changes: ', e);
                this.dbConnection = null;
            }
        } else {
            try {
                await this.dbConnection.ping();
            } catch(e) {
                logger.error('Ping Error: ', e);
            }
        }
    }
}
