import { State } from '../state/index.mjs';
import { WS_CHANNELS, ZMQ_ACTION, ZMQ_OBJECT } from '../../system/index.mjs';

export class Notify {
    /**
     * @class {Notify}
     * */

    static envelope(message, channel) {
        return {
            jsonrpc: '2.0',
            method: 'notify',
            params: {
                subscription: channel,
                data: { ...message }
            }
        }
    }

    /**
     * @method
     * @param {Object} positionRaw  - raw position form JSON/BIN convertor
     * @param {String} positionRaw.ts  - position timestamp in ISO 8601 format
     * @param {String} positionRaw.tk_uid  - trackable object uid
     * @param {Object} positionRaw.data  - position properties
     * @param {Number} positionRaw.data.x  - the X coordinate of the position
     * @param {Number} positionRaw.data.y  - the Y coordinate of the position
     * @param {Number} positionRaw.data.z  - the Z coordinate of the position
     * @param {String} positionRaw.data.type  - positions type (UWB, HB)
     * @param {String} positionRaw.data.sl_uid  - sublocation numeric identifier
     * @param {String} positionRaw.data.mac  - hardware MAC address
     * @returns {Promise<{topic: string|null, message: Object}>}
     */
    static async getPositionMessage(positionRaw) {
        const mac = (positionRaw.data.mac || '').toUpperCase();
        const macAssociation = await new State().readNodeProject(mac);
        if (!macAssociation) return { message: positionRaw, topic: null };

        const timestamp = new Date(positionRaw.ts).getTime();
        const ms = timestamp % 1000;
        const ts = (timestamp - ms) / 1000;

        const message = {
            ts,
            ms,
            dev_id: positionRaw.tk_uid || 'unassigned',
            id: null, // marked as deprecated in public API
            mac,
            x: positionRaw.data.x,
            y: positionRaw.data.y,
            z: positionRaw.data.z,
            locationId: null, // marked as deprecated in public API
            sl: null, // marked as deprecated in public API
            sl_id: positionRaw.data.sl_uid || '',
            fom: positionRaw.data.fom || 0,
            er: null,
            type: positionRaw.data.type,
            to_type_id: null, // marked as deprecated in public API
            sfsn: null, // no value in a platform
            slvr_mode: null, // no value in a platform
            pe_dim: null, // no value in a platform
            inactive: null, // no value in a platform
            edge_corr: null, // no value in a platform
            project_uid: macAssociation.projectUid,
            company_uid: macAssociation.companyUid,
        }
        const topic = `${WS_CHANNELS.POS}:${macAssociation.projectUid}`;

        return {
            topic,
            message: Notify.envelope(message, WS_CHANNELS.POS),
        }
    }
    
    /**
     * Get Project uid from #notify message. Default value is in notifyRaw.project_uid
     * @param {Object} notifyRaw
     * @property {String} notifyRaw.object - ZMQ objects:['WIN'|'USER'|'PROJECT_USER_ACCESS',...]
     * @property {String} notifyRaw.action - ZMQ actions: ['created'|'updated'|'deleted'|'release'|'assign'].
     * @property {String} notifyRaw.mac - win mac_address used for object 'WIN'
     * @returns {Promise<string|*>}
     */
    static async getProjectUid(notifyRaw) {
        switch (notifyRaw.object) {
            case ZMQ_OBJECT.WIN: {
                const node = await (new State()).readNodeProject(notifyRaw.mac);
                return node && node.projectUid;
            }
            case ZMQ_OBJECT.USER:
            case ZMQ_OBJECT.PROJECT_USER_ACCESS:
            default: {
                return notifyRaw.project_uid
            }
        }
    }

    /**
     * @method
     * @param {Object} notifyRaw - notify channel message
     * @param {Number} notifyRaw.ts - unix timestamp
     * @param {Number} notifyRaw.ms - milliseconds part of the timestamp
     * @param {String|undefined} notifyRaw.project_uid - project unique identifier
     * @param {String|undefined} notifyRaw.company_uid - company unique identifier
     * @param {String} notifyRaw.mac - device mac address
     * @param {String} notifyRaw.object - entity name
     * @param {String} notifyRaw.action - action name
     * @param {Object} notifyRaw.plist - payload
     * @return {Object<{topic: string, message: Object}>}
     */
    static async getNotifyMessage(notifyRaw) {
        const projectUid = await Notify.getProjectUid(notifyRaw);
        
        const topic = `${WS_CHANNELS.NOTIFY}:${projectUid}`;

        return {
            topic,
            message: Notify.envelope(notifyRaw, WS_CHANNELS.NOTIFY),
        }
    }

    /**
     * @method
     * @param {Object} alarmRaw - alarm channel message
     * @param {String} alarmRaw.project_uid - project unique identifier
     * @return {Object<{topic: string, message: Object}>}
     */
    static getAlarmMessage(alarmRaw) {
        const topic = `${WS_CHANNELS.ALARM}:${alarmRaw.project_uid}`;

        return {
            topic,
            message: Notify.envelope(alarmRaw, WS_CHANNELS.ALARM),
        }
    }

    /**
     * @method
     * @param {Object} sensorRawData - sensor channel message
     * @param {String} sensorRawData.company_uid - company unique identifier
     * @param {String} sensorRawData.project_uid - project unique identifier
     * @return {Object<{topic: string, message: Object}>}
     */
    static getSensorMessage(sensorRawData) {
        const topic = `${WS_CHANNELS.SENSOR}:${sensorRawData.project_uid}`;

        return {
            topic,
            message: Notify.envelope(sensorRawData, WS_CHANNELS.SENSOR),
        }
    }

    /**
     * @method
     * @param {Object} otaRawData - sensor channel message
     * @param {String} otaRawData.mac - hardware MAC address
     * @return {Promise<{topic: string|null, message: Object}>}
     */
    static async getOtaMessage(otaRawData) {
        const mac = (otaRawData.mac || '').toUpperCase();
        const macAssociation = await new State().readNodeProject(mac);
        if (!macAssociation) return { message: otaRawData, topic: null };
        const topic = `${WS_CHANNELS.OTA}:${macAssociation.projectUid}`;

        return {
            topic,
            message: Notify.envelope(otaRawData, WS_CHANNELS.OTA),
        }
    }

    /**
     * @method
     * @param {Object} zoneEventRaw - raw zone event
     * @param {String} zoneEventRaw.action - zone event action (enter, leave)
     * @param {Number} zoneEventRaw.ts - unix timestamp
     * @param {Number} zoneEventRaw.ms - millisecond part of te timestamp
     * @param {String} zoneEventRaw.obj_uid - trackable object uid
     * @param {String} zoneEventRaw.project_uid - project uid
     * @param {String} zoneEventRaw.sublocation_uid - sublocation uid
     * @param {Number} zoneEventRaw.x - zone event X coordinate
     * @param {Number} zoneEventRaw.y - zone event Y coordinate
     * @param {Number} zoneEventRaw.z - zone event Z coordinate
     * @param {String} zoneEventRaw.zone_type_name - zone type
     * @param {String} zoneEventRaw.zone_uid - zone unique identifier
     * @returns {Object<{topic: string, message: Object}>}
     */
    static getZoneEventMessage(zoneEventRaw) {
        const topic = `${WS_CHANNELS.ZONE_EVENT}:${zoneEventRaw.project_uid}`;

        return {
            topic,
            message: Notify.envelope(zoneEventRaw, WS_CHANNELS.ZONE_EVENT),
        }
    }

    /**
     * @method
     * @param {Object} paramPRaw - raw param_p message
     * @param {String} paramPRaw.mac - hardware MAC address
     * @param {Number} paramPRaw.ts - unix timestamp
     * @param {Number} paramPRaw.ms - milliseconds part of the timestamp
     * @param {Object} paramPRaw.plist - message payload
     * @returns {Promise<{topic: string|null, message: Object}>}
     */
    static async getParamPMessage(paramPRaw) {
        const macAssociation = await new State().readNodeProject(paramPRaw.mac);
        if (!macAssociation) return { message: paramPRaw, topic: null };
        const topic = `${WS_CHANNELS.PARAM_P}:${macAssociation.projectUid}`;

        return {
            topic,
            message: Notify.envelope(paramPRaw, WS_CHANNELS.PARAM_P),
        }
    }
}
