import { State } from './State.mjs';
import { SubscriptionsManager } from '../subscription/index.mjs';
import { ZMQ_ACTION, ZMQ_OBJECT } from '../../system/index.mjs';


export class StateManager {
    /**
     * @class {StateManager}
     * */

    /**
     * @method
     * @param {Object} message - ZMQ notify channel message
     * @param {String} message.action - action name (created, updated, deleted)
     * @param {String} message.object - object name (created, updated, deleted)
     * @param {Object} message.plist - message payload
     * @param {String} message.project_uid - project unique identifier
     * @param {String} message.company_uid - company unique identifier
     * @return {Object}
     */
    static format(message) {
        return {
            mac: message.plist.mac_address,
            projectUid: message.project_uid,
            companyUid: message.company_uid
        }
    }

    /**
     * @method
     * @param {Object} message - ZMQ notify channel message
     * @param {String} message.object - entity name
     * @param {String} message.action - action name (created, updated, deleted)
     * @param {Object} message.plist - message payload
     * @param {String} message.project_uid - project unique identifier
     * @param {String} message.company_uid - company unique identifier
     * @return {undefined}
     */
    static async handleNotifyMessage(message) {
        switch (message.object) {
            case(ZMQ_OBJECT.TRACKABLE_OBJECT_ASSOCIATION):
            case(ZMQ_OBJECT.WIN_ANCHOR_ASSOCIATION): {
                await StateManager.handleAssociationMessage(message);
                break;
            }
            case(ZMQ_OBJECT.NODE): {
                await StateManager.handleNodeMessage(message);
                break;
            }
            case(ZMQ_OBJECT.USER):
            case(ZMQ_OBJECT.PROJECT_USER_ACCESS): {
                StateManager.handleUserMessage(message);
                break;
            }
        }

    }

    /**
     * @method
     * @param {Object} message - ZMQ notify channel message
     * @param {String} message.object - entity name
     * @param {String} message.action - action name (created, updated, deleted)
     * @param {Object} message.plist - message payload
     * @param {String} message.project_uid - project unique identifier
     * @param {String} message.company_uid - company unique identifier
     * @return {Promise<undefined>}
     */
    static async handleAssociationMessage(message) {
        if (!message.project_uid) return undefined;
        const formattedMsg = StateManager.format(message);
        switch (message.action) {
            case ZMQ_ACTION.CREATED: {
                await new State().saveNodeProject(formattedMsg);
                break;
            }
            case ZMQ_ACTION.DELETED:
            case ZMQ_ACTION.UPDATED: {
                // Association closing not meant that node is released from the project
                await new State().saveNodeProject(formattedMsg);
                break;
            }
        }
    }

    /**
     * @method
     * @param {Object} message - ZMQ notify channel message
     * @param {String} message.object - entity name
     * @param {String} message.action - action name (created, updated, deleted)
     * @param {Object} message.plist - message payload
     * @param {String} message.project_uid - project unique identifier
     * @param {String} message.company_uid - company unique identifier
     * @return {Promise<undefined>}
     */
    static async handleNodeMessage(message) {
        const formattedMsg = StateManager.format(message);
        switch (message.action) {
            case ZMQ_ACTION.CREATED:
            case ZMQ_ACTION.UPDATED: {
                if (message.project_uid) {
                    await new State().saveNodeProject(formattedMsg);
                }
                break;
            }
            case ZMQ_ACTION.RELEASE: {
                await new State().removeNodeProject(formattedMsg);
                break;
            }
            case ZMQ_ACTION.ASSIGN: {
                if (message.project_uid) {
                    await new State().saveNodeProject(formattedMsg);
                }
                break;
            }
            case ZMQ_ACTION.DELETED: {
                await new State().removeNodeProject(formattedMsg);
                break;
            }
        }
    }

    /**
     * @method
     * @param {Object} message - ZMQ notify channel message
     * @param {String} message.object - entity name
     * @param {String} message.action - action name (created, updated, deleted)
     * @param {Object} message.plist - message payload
     * @param {String} message.project_uid - project unique identifier
     * @param {String} message.company_uid - company unique identifier
     * @return {undefined}
     */
    static handleUserMessage(message) {
        switch (message.action) {
            case ZMQ_ACTION.DELETED: {
                const { deletedUserEmail } = message.plist;
                const subscriptionsManager = new SubscriptionsManager();
                subscriptionsManager.closeUserClients(deletedUserEmail);
                break;
            }
            case ZMQ_ACTION.UPDATED: {
                const { email } = message.plist;
                const subscriptionsManager = new SubscriptionsManager();
                subscriptionsManager.closeUserClients(email);
                break;
            }
        }
    }
}
