import { SubscriptionsManagerBase } from './SubscriptionsManagerBase.mjs';

export class SubscriptionsManager extends SubscriptionsManagerBase {
    /**
     * @typedef {Class} SubscriptionsManager
     * @method findSubscriptionsByEmailClientUid
     * @method findClientsByEmail
     * @method findSubscribedClientByParams
     * @method findSubscribersByTopic
     * @method addSubscription
     * @method deleteSubscriber
     * @method deleteSubscriptionByEmail
     * @method deleteSubscriptionByParams
     * @method closeUserClients
     */

    constructor(props) {
        super(props);
    }

    /**
     * @method
     * @returns {Array<{topic: string, email: string, clientsNo: number}>}
     */
    getAllSubscriptions() {
        const subscriptions = [];
        for (const topic in this.subscriptions) {
            const subscribedUsers = this.subscriptions[topic] && Object.keys(this.subscriptions[topic]);
            subscribedUsers.forEach(email => {
                const clientsNo = (this.subscriptions[topic][email] || []).length;
                subscriptions.push({ topic, email, clientsNo });
            })
        }
        return subscriptions;
    }

    /**
     * @param {string} email - subscribed user email
     * @param {string} clientUid - client unique identifier
     * @returns {Array<{topic: string, client: Object}>}
     */
    findSubscriptionsByEmailClientUid(email, clientUid) {
        const subscriptions = [];
        for (const topic in this.subscriptions) {
            const subscribedClients = this.subscriptions[topic] && this.subscriptions[topic][email] || [];
            const client = subscribedClients.find(sub => sub.client.uid === clientUid);
            if (client) {
                subscriptions.push({ topic, client });
            }
        }

        return subscriptions;
    }

    /**
     * @param {string} email - subscribed user email
     * @returns {Object}
     */
    findClientsByEmail(email) {
        const subscriptions = {};
        for (const topic in this.subscriptions) {
            const clients = this.subscriptions[topic] && this.subscriptions[topic][email] || [];
            clients.forEach(client => {
                subscriptions[client.uid] = client;
            });
        }
        return subscriptions;
    }

    /**
     * @param {string} topic - subscription topic
     * @param {string} email - subscribed user email
     * @param {string} clientUid - client unique identifier
     * @returns {ClientBase}
     */
    findSubscribedClientByParams(topic, email, clientUid) {
        if (!this.subscriptions[topic] || !this.subscriptions[topic][email]) return null;
        return this.subscriptions[topic][email].find(subscription => (subscription.client && subscription.client.uid === clientUid));
    }

    /**
     * @param {string} topic - subscription topic
     * @returns {Object}
     */
    findSubscribersByTopic(topic) {
        return this.subscriptions && this.subscriptions[topic];
    }

    /**
     * @param {ClientBase} client
     * @param {string} topic
     * @param {string} clientUid
     * @returns {undefined}
     */
    addSubscription(client, topic, clientUid) {
        const clients = this.subscriptions[topic];
        if (!clients) {
            this.subscriptions[topic] = {[client.email]: [client]};
            // Set callback on client close for client onClose method WsClient.mjs
            // Cleanup subscription on client closing
            client.setOnClose(this.deleteSubscriptionByEmail.bind(this));
            return undefined;
        }

        if (!clients[client.email]) {
            this.subscriptions[topic][client.email] = [client];
            // Set callback on client close for client onClose method WsClient.mjs
            // Cleanup subscription on client closing
            client.setOnClose(this.deleteSubscriptionByEmail.bind(this));
            return undefined;
        }

        // Looking for subscribed user client
        const subscribedClient = clients[client.email].find(client => client.uid === clientUid);
        // In case subscription already exist do nothing
        if (!subscribedClient) {
            // Add a new client for a user
            this.subscriptions[topic][client.email].push(client);
            // Set callback on client close for client onClose method WsClient.mjs
            // Cleanup subscription on client closing
            client.setOnClose(this.deleteSubscriptionByEmail.bind(this));
            return undefined;
        }
        return undefined;
    }

    /**
     * @param {string} topic
     * @param {string} email
     * @param {string} clientUid
     * @returns {undefined}
     */
    deleteSubscriber(topic, email, clientUid) {
        const existedSubscription = this.findSubscribedClientByParams(topic, email, clientUid);
        if (existedSubscription) {
            this.subscriptions[topic][email] = this.subscriptions[topic][email].filter(sub => sub.client.uid !== clientUid);
            if (!Object.keys(this.subscriptions[topic][email]).length) {
                delete this.subscriptions[topic][email];
            }
        }

        if (this.subscriptions[topic] && !Object.keys(this.subscriptions[topic]).length) {
            delete this.subscriptions[topic];
        }
    }

    /**
     * @param {string} email - users email
     * @param {string} clientUid - client unique identifier
     * @returns {undefined}
     */
    deleteSubscriptionByEmail(email, clientUid) {
        const subscriptions = this.findSubscriptionsByEmailClientUid(email, clientUid);
        for (const subscription of subscriptions) {
            const email = subscription.client && subscription.client.email;
            const clientUid = subscription.client && subscription.client.uid;
            const topic = subscription.topic;

            this.deleteSubscriber(topic, email, clientUid);
        }
    }

    /**
     * @param {string} email - users email
     * @param {Array<string>} channels - channels list
     * @param {string} project - project unique identifier
     * @param {string} clientUid - unique identifier of the client
     * @returns {undefined}
     */
    deleteSubscriptionByParams(email, channels, project, clientUid) {
        for (const channel of channels) {
            const topic = `${channel}:${project}`;
            this.deleteSubscriber(topic, email, clientUid);
        }
    }

    /**
     * @param {string} email - users email
     * @returns {undefined}
     */
    closeUserClients(email) {
        const clients = this.findClientsByEmail(email);
        // Close connection. Will trigger onClose that will clear all subscriptions for a client
        for (const clientUid in clients) {
            if (clients[clientUid]) {
                clients[clientUid].close()
            }
        }
    }
}
