import assert from 'assert';
import { SubscriptionsManager } from './SubscriptionsManager.mjs';
import { SubscriptionsManagerBase } from './SubscriptionsManagerBase.mjs';
import { WsClient } from '../client/index.mjs';
import { generateUid } from '../../utils/index.mjs';

const email = 'test@redpointpositioning.com';
const projectUid = generateUid();
let clientUidOne = generateUid();
let clientUidTwo = generateUid();
function generateWsClient(uid) {
    return {
        uid,
        on: () => {}
    }
}

describe('SubscriptionsManager', function () {
    before(function () {
        SubscriptionsManagerBase.init();
    })

    it('should add one subscription', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const client = new WsClient(generateWsClient(clientUidOne), email, clientUidOne);
        const topic = `${projectUid}:pos`;

        subscriptionsManager.addSubscription(client, topic, clientUidOne);

        assert.deepStrictEqual(subscriptionsManager.subscriptions[topic], {[email]: [client]}, `not equal`);
    });

    it('should add one more subscription', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const client = new WsClient(generateWsClient(clientUidTwo), email, clientUidTwo);
        const topic = `${projectUid}:pos`;

        subscriptionsManager.addSubscription(client, topic, clientUidTwo);

        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions), [topic]);
        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions[topic]), [email]);
        assert.strictEqual(subscriptionsManager.subscriptions[topic][email].length, 2, `not equal`);
    });

    it('should add subscription to alarm channel', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const client = new WsClient(generateWsClient(clientUidOne), email, clientUidOne);
        const topicOne = `${projectUid}:pos`;
        const topicTwo = `${projectUid}:alarm`;

        subscriptionsManager.addSubscription(client, topicTwo, clientUidOne);

        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions), [topicOne, topicTwo]);
        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions[topicTwo]), [email]);
        assert.strictEqual(subscriptionsManager.subscriptions[topicTwo][email].length, 1, `not equal`);
    });

    it('should delete subscription to alarm channel', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const topicOne = `${projectUid}:pos`;
        const topicTwo = `${projectUid}:alarm`;

        subscriptionsManager.deleteSubscriber(topicTwo, email, clientUidOne);

        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions), [topicOne]);
        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions[topicOne]), [email]);
        assert.strictEqual(subscriptionsManager.subscriptions[topicOne][email].length, 2, `not equal`);
    });

    it('should delete subscription for email client uid two', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const topicOne = `${projectUid}:pos`;

        subscriptionsManager.deleteSubscriptionByEmail(email, clientUidTwo);

        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions), [topicOne]);
        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions[topicOne]), [email]);
        assert.strictEqual(subscriptionsManager.subscriptions[topicOne][email].length, 1, `not equal`);
    });

    it('should delete subscription for email client uid one', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        subscriptionsManager.deleteSubscriptionByEmail(email, clientUidOne);
        assert.deepStrictEqual(Object.keys(subscriptionsManager.subscriptions), []);
    });

    it('should find subscribers by topic', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const client = new WsClient(generateWsClient(clientUidOne), email, clientUidOne);
        const topicOne = `${projectUid}:pos`;

        subscriptionsManager.addSubscription(client, topicOne, clientUidOne);
        const subscriber = subscriptionsManager.findSubscribersByTopic(topicOne);

        assert.deepStrictEqual(subscriber[email].length, 1);
        assert.deepStrictEqual(subscriber[email][0].uid, clientUidOne);
    });

    it('should find clients by email', async function() {
        const subscriptionsManager = new SubscriptionsManager();
        const clients = subscriptionsManager.findClientsByEmail(email);
        assert.deepStrictEqual(Object.keys(clients), [clientUidOne]);
    });
    // todo: cover with the test cases other methods
});
